from __future__ import annotations

import dataclasses
import math
import time
import typing
import uuid

import requests
from pantos.common.blockchains.base import Blockchain
from pantos.common.entities import ServiceNodeBid
from pantos.common.entities import \
    TokenDeploymentRequest as TokenDeploymentSubmissionRequest
from pantos.common.types import BlockchainAddress
from pantos.common.types import PrivateKey

from pantos.client.library.blockchains import BlockchainClient
from pantos.client.library.blockchains import get_blockchain_client
from pantos.client.library.business.base import Interactor
from pantos.client.library.business.base import InteractorError
from pantos.client.library.configuration import config
from pantos.client.library.configuration import get_blockchain_config

_DEPLOYMENT_RESOURCE = 'deployment'

_PAYMENT_RESOURCE = 'payment'

_CHEAPEST_BID_RESOURCE = 'bids/cheapest'

_DEFAULT_VALID_UNTIL_BUFFER = 10000000


class TokenDeploymentInteractorError(InteractorError):
    """Exception class for all deployment interactor errors.

    """
    pass


class TokenDeploymentInteractor(Interactor):
    """Interactor for handling Pantos compatible token deployments.

    """
    @dataclasses.dataclass
    class TokenDeploymentRequest:
        """Request data for a new token deployment.

        Attributes
        ----------
        token_name : str
            The name of the token.
        token_symbol : str
            The symbol of the token.
        token_decimals : int
            The token's number of decimals.
        token_pausable : bool
            If the token is pausable.
        token_burnable : bool
            If the token is burnable.
        token_supply : int
            The supply of the token.
        deployment_blockchains : list of Blockchain
            The blockchains where the deployment will be requested.
        payment_blockchain : Blockchain
            The blockchain on which the payment for the fee will be made.
        payer_private_key : PrivateKey
            The unencrypted private key of the payer's account on the
            payment_blockchain.
        valid_until_buffer : int
            The buffer in seconds added to the current timestamp plus
            the chosen service node bid's execution time.

        """
        token_name: str
        token_symbol: str
        token_decimals: int
        token_pausable: bool
        token_burnable: bool
        token_supply: int
        deployment_blockchains: typing.List[Blockchain]
        payment_blockchain: Blockchain
        payer_private_key: PrivateKey
        valid_until_buffer: int = _DEFAULT_VALID_UNTIL_BUFFER

    @dataclasses.dataclass
    class __DeploymentFee:
        """Deployment fee data for deploying Pantos compatible
        token smart contracts.

        Attributes
        ----------
        amount : int
            The amount to be paid.
        symbol : str
            The token symbol.

        """
        amount: int
        symbol: str

    @dataclasses.dataclass
    class __PaymentResponse:
        """Response data when requesting the payment from the
        Pantos token creator service.

        Attributes
        ----------
        receiver_address : str
            The receiver_address.
        deployment_fee_valid_until : int
            The timestamp until when the payment can be submitted
            to the token creator service (in seconds since the epoch).
        signature : str
            The signature that ensures the authenticity and valability
            of the payment response.
        deployment_fee : TokenDeploymentInteractor.__DeploymentFee
            The deployment fee object.

        """
        receiver_address: str
        deployment_fee_valid_until: int
        signature: str
        deployment_fee: TokenDeploymentInteractor.__DeploymentFee

    def deploy_token(self, request: TokenDeploymentRequest) -> uuid.UUID:
        """Deploy a Pantos compatible token.

        Parameters
        ----------
        request : TokenDeploymentRequest
            The request data for a new token deployment.

        Returns
        -------
        uuid.UUID
            The Pantos token creator's task ID.

        Raises
        ------
        TokenDeploymentInteractorError
            If the token deployment cannot be executed.

        """
        deployment_blockchain_ids = [
            blockchain.value for blockchain in request.deployment_blockchains
        ]
        service_node_address, service_node_bid = \
            self.__get_cheapest_bid_response(request.payment_blockchain)
        valid_until = self.__compute_valid_until(
            service_node_bid.execution_time, request.valid_until_buffer)
        blockchain_config = get_blockchain_config(request.payment_blockchain)
        pan_token_address = blockchain_config['tokens']['pan']
        source_blockchain_client = get_blockchain_client(
            request.payment_blockchain)
        payment_response = self.__get_payment_response(
            request.payment_blockchain, request.deployment_blockchains)
        recipient_address = BlockchainAddress(
            payment_response.receiver_address)
        transfer_signature_request = \
            BlockchainClient.ComputeTransferSignatureRequest(
                request.payer_private_key, recipient_address,
                pan_token_address,
                payment_response.deployment_fee.amount,
                service_node_address, service_node_bid, valid_until)
        transfer_signature_response = \
            source_blockchain_client.compute_transfer_signature(
                    transfer_signature_request)
        deployment_request = TokenDeploymentSubmissionRequest(
            deployment_blockchain_ids, request.token_name,
            request.token_symbol, request.token_decimals,
            request.token_pausable, request.token_burnable,
            request.token_supply, request.payment_blockchain.value,
            transfer_signature_response.sender_address,
            payment_response.deployment_fee.amount, payment_response.signature,
            payment_response.deployment_fee_valid_until, service_node_bid.fee,
            service_node_bid.execution_time, service_node_bid.valid_until,
            service_node_bid.signature,
            transfer_signature_response.sender_nonce, valid_until,
            transfer_signature_response.signature)
        return self.__submit_deployment_request(deployment_request)

    def __submit_deployment_request(
            self, request: TokenDeploymentSubmissionRequest) -> uuid.UUID:
        """Submit a deployment request to the Pantos token creator service.

        Parameters
        ----------
        request : TokenDeploymentSubmissionRequest
            Request data for submitting a new deployment request to the
            Pantos token creator service.

        Returns
        -------
        uuid.UUID
            The Pantos token creator's task ID.

        Raises
        ------
        TokenDeploymentInteractorError
            If the token deployment request cannot be submitted
            successfully.

        """
        payload = {
            'deployment_blockchain_ids': request.deployment_blockchain_ids,
            'token_name': request.token_name,
            'token_symbol': request.token_symbol,
            'token_decimals': request.token_decimals,
            'token_pausable': request.token_pausable,
            'token_burnable': request.token_burnable,
            'token_supply': request.token_supply,
            'payment_blockchain_id': request.payment_blockchain_id,
            'payer_address': request.payer_address,
            'deployment_fee': request.deployment_fee,
            'deployment_fee_valid_until': request.deployment_fee_valid_until,
            'deployment_fee_signature': request.deployment_fee_signature,
            'bid': {
                'fee': request.bid_fee,
                'execution_time': request.bid_execution_time,
                'valid_until': request.bid_valid_until,
                'signature': request.bid_signature
            },
            'payment_nonce': request.payment_nonce,
            'payment_valid_until': request.payment_valid_until,
            'payment_signature': request.payment_signature
        }
        token_creator_url_deployment = self.__build_resource_url(
            config['token_creator']['url'], _DEPLOYMENT_RESOURCE)
        try:
            token_creator_response = requests.post(
                token_creator_url_deployment, json=payload)
            token_creator_response.raise_for_status()
            task_id = token_creator_response.json()['task_id']
            return uuid.UUID(task_id)
        except (requests.exceptions.HTTPError, KeyError):
            raise TokenDeploymentInteractorError(
                'unable to submit a new deployment request', extra=request)

    def __get_payment_response(
            self, payment_blockchain: Blockchain,
            deployment_blockchains: typing.List[Blockchain]) \
            -> __PaymentResponse:
        """Get the payment response from the Pantos token
        creator service.

        Parameters
        ----------
        payment_blockchain : int
            The blockchain where the payment will be done.
        deployment_blockchains : list of Blockchain
            The list of deployment blockchains.

        Returns
        -------
        __PaymentResponse
            Response data when requesting the payment from the
            Pantos token creator service.

        Raises
        ------
        TokenDeploymentInteractorError
            If the token deployment request cannot be submitted
            successfully.

        """
        payment_blockchain_id = payment_blockchain.value
        deployment_blockchain_ids = [
            blockchain.value for blockchain in deployment_blockchains
        ]
        token_creator_url_deployment_fee = self.__build_resource_url(
            config['token_creator']['url'], _PAYMENT_RESOURCE)
        request = {
            'payment_blockchain_id': payment_blockchain_id,
            'deployment_blockchain_ids': deployment_blockchain_ids
        }
        try:
            token_creator_response = requests.post(
                token_creator_url_deployment_fee, json=request)
            token_creator_response.raise_for_status()
            json_response = token_creator_response.json()
            deployment_fee = json_response['fee']
            return self.__PaymentResponse(
                json_response['receiver_address'],
                json_response['valid_until'], json_response['signature'],
                self.__DeploymentFee(int(deployment_fee['amount']),
                                     deployment_fee['symbol']))
        except (requests.exceptions.HTTPError, KeyError):
            raise TokenDeploymentInteractorError(
                'unable to submit a new deployment request', extra=request)

    def __get_cheapest_bid_response(
            self, payment_blockchain: Blockchain) \
            -> typing.Tuple[BlockchainAddress, ServiceNodeBid]:
        """Get the cheapest bid response from the Pantos token
        creator service.

        Parameters
        ----------
        payment_blockchain : int
            The blockchain where the payment will be done.

        Returns
        -------
        Tuple[BlockchainAddress, ServiceNodeBid]
            A tuple containing the blockchain address of the service node with
            it's cheapest bid.

        Raises
        ------
        TokenDeploymentInteractorError
            If unable to get the cheapest bid response.

        """
        payment_blockchain_id = payment_blockchain.value
        token_creator_cheapest_bid_url = self.__build_resource_url(
            config['token_creator']['url'], _CHEAPEST_BID_RESOURCE)
        query_parameters = f'payment_blockchain_id={payment_blockchain_id}'
        try:
            token_creator_response = requests.get(
                f'{token_creator_cheapest_bid_url}?{query_parameters}')
            token_creator_response.raise_for_status()
            json_response = token_creator_response.json()
            service_node_bid = ServiceNodeBid(
                source_blockchain=payment_blockchain,
                destination_blockchain=payment_blockchain,
                fee=json_response['fee'],
                execution_time=json_response['execution_time'],
                valid_until=json_response['valid_until'],
                signature=json_response['signature'])
            return json_response['service_node_address'], service_node_bid
        except (requests.exceptions.HTTPError, KeyError):
            raise TokenDeploymentInteractorError(
                'unable to submit a cheapets bid request')

    def __compute_valid_until(self, execution_time: int,
                              valid_until_buffer: int) -> int:
        """Compute the valid until buffer for the payment.

        Parameters
        ----------
        execution_time : int
            The execution time of the payment.
        valid_until_buffer : int
            The buffer for sending the payment.

        Returns
        -------
        int
            The timestamp until when the payment can be submitted
            to the service node (in seconds since the epoch).

        """
        if valid_until_buffer < 0:
            raise TokenDeploymentInteractorError(
                '"valid until" buffer must be non-negative',
                valid_until_buffer=valid_until_buffer)
        return (math.ceil(time.time()) + execution_time + valid_until_buffer)

    def __build_resource_url(self, token_creator_url: str,
                             resource: str) -> str:
        """Build the token creator resource url.

        Parameters
        ----------
        token_creator_url : str
            The token creator url.

        Returns
        -------
        str
            The token creator resource url.

        """
        transfer_url = token_creator_url
        if not token_creator_url.endswith('/'):
            transfer_url += '/'
        transfer_url += resource
        return transfer_url



"""Module for loading, validating, and accessing the client library's
configuration.

"""
import itertools
import typing

from pantos.common.blockchains.base import Blockchain
from pantos.common.configuration import Config

_DEFAULT_FILE_NAME: typing.Final[str] = 'client-library.yml'
"""Default configuration file name."""

config = Config(_DEFAULT_FILE_NAME)
"""Singleton object holding the configuration values."""

_VALIDATION_SCHEMA_BLOCKCHAIN = {
    'type': 'dict',
    'schema': {
        'active': {
            'type': 'boolean',
            'default': True
        },
        'provider': {
            'type': 'string',
            'required': True
        },
        'fallback_providers': {
            'type': 'list',
            'required': False,
            'schema': {
                'type': 'string',
            }
        },
        'average_block_time': {
            'type': 'integer',
            'required': True
        },
        'blocks_per_query': {
            'type': 'integer',
            'required': True,
            'min': 1
        },
        'chain_id': {
            'type': 'integer',
            'required': True
        },
        'confirmations': {
            'type': 'integer',
            'required': True
        },
        'hub': {
            'type': 'string',
            'required': True
        },
        'forwarder': {
            'type': 'string',
            'required': True
        },
        'tokens': {
            'type': 'dict',
            'required': True,
            'keysrules': {
                'type': 'string',
                'regex': '[a-z0-9]+'
            },
            'valuesrules': {
                'type': 'string'
            }
        }
    }
}
"""Schema for validating a blockchain entry in the configuration file."""

_VALIDATION_SCHEMA = {
    'protocol': {
        'type': 'dict',
        'required': True,
        'schema': {
            'mainnet': {
                'type': 'string',
                'required': True,
                'regex': r'^[0-9]+\.[0-9]+\.[0-9]+$'
            },
            'testnet': {
                'type': 'string',
                'required': True,
                'regex': r'^[0-9]+\.[0-9]+\.[0-9]+$'
            }
        }
    },
    'token_creator': {
        'type': 'dict',
        'schema': {
            'url': {
                'type': 'string',
                'required': True
            }
        }
    },
    'service_nodes': {
        'type': 'dict',
        'schema': {
            'timeout': {
                'type': 'float',
                'required': True
            }
        }
    },
    'blockchains': {
        'type': 'dict',
        'schema': dict(
            zip([b.name.lower() for b in Blockchain],
                itertools.repeat(_VALIDATION_SCHEMA_BLOCKCHAIN)))
    }
}
"""Schema for validating the configuration file."""


def get_blockchain_config(
        blockchain: Blockchain) -> typing.Dict[str, typing.Any]:
    """Get a blockchain-specific configuration dictionary.

    Parameters
    ----------
    blockchain : Blockchain
        The blockchain to get the configuration for.

    Returns
    -------
    dict
        The blockchain-specific configuration.

    """
    return config['blockchains'][blockchain.name.lower()]


def load_config(file_path: typing.Optional[str] = None,
                reload: bool = True) -> None:
    """Load the configuration from a configuration file.

    Parameters
    ----------
    file_path : str or None
        The path to the configuration file (typical configuration file
        locations are searched if none is specified).
    reload : bool
        If True, the configuration is also loaded if it was already
        loaded before.

    Raises
    ------
    pantos.common.configuration.ConfigError
        If the configuration cannot be loaded (e.g. due to an invalid
        configuration file).

    See Also
    --------
    Config.load

    """
    if reload or not config.is_loaded():
        config.load(_VALIDATION_SCHEMA, file_path)



"""Base classes for all business logic interactors and errors.

"""
import abc

from pantos.client.library.exceptions import ClientLibraryError


class InteractorError(ClientLibraryError):
    """Base exception class for all interactor errors.

    """
    pass


class Interactor(abc.ABC):
    """Base class for all interactors.

    """
    pass



"""Business logic for handling Pantos token transfers.

"""
import dataclasses
import math
import time
import typing
import uuid

from pantos.common.blockchains.base import Blockchain
from pantos.common.entities import BlockchainAddressBidPair
from pantos.common.entities import ServiceNodeBid
from pantos.common.entities import ServiceNodeTransferStatus
from pantos.common.servicenodes import ServiceNodeClient
from pantos.common.types import Amount
from pantos.common.types import BlockchainAddress
from pantos.common.types import PrivateKey
from pantos.common.types import TokenId

from pantos.client.library.blockchains import BlockchainClient
from pantos.client.library.blockchains import get_blockchain_client
from pantos.client.library.blockchains.base import UnknownTransferError
from pantos.client.library.business.base import Interactor
from pantos.client.library.business.base import InteractorError
from pantos.client.library.business.bids import BidInteractor
from pantos.client.library.business.tokens import TokenInteractor
from pantos.client.library.configuration import get_blockchain_config
from pantos.client.library.entitites import DestinationTransferStatus
from pantos.client.library.entitites import ServiceNodeTaskInfo
from pantos.client.library.entitites import TokenTransferStatus

_DEFAULT_VALID_UNTIL_BUFFER = 120
"""Default "valid until" timestamp buffer for a token transfer in seconds."""


class TransferInteractorError(InteractorError):
    """Exception class for all transfer interactor errors.

    """
    pass


class TransferInteractor(Interactor):
    """Interactor for handling Pantos token transfers.

    """
    @dataclasses.dataclass
    class TokenTransferStatusRequest:
        """Request data for the status of a token transfer.

        Attributes
        ----------
        source_blockchain : Blockchain
            The token transfer's source blockchain.
        service_node_address : BlockchainAddress
            The address of the service node that is processing the
            token transfer.
        service_node_task_id : uuid.UUID
            The service node task ID of the token transfer.
        blocks_to_search : int or None
            The number of blocks to search for the token transfer
            (default: None).

        """
        source_blockchain: Blockchain
        service_node_address: BlockchainAddress
        service_node_task_id: uuid.UUID
        blocks_to_search: int | None = None

    @dataclasses.dataclass
    class TransferTokensRequest:
        """Request data for a new token transfer.

        Attributes
        ----------
        source_blockchain : Blockchain
            The token transfer's source blockchain.
        destination_blockchain : Blockchain
            The token transfer's destination blockchain.
        sender_private_key : PrivateKey
            The unencrypted private key of the sender's account on the
            source blockchain.
        recipient_address : BlockchainAddress
            The address of the recipient's account on the destination
            blockchain.
        source_token_id : TokenId
            The identifier of the token to be transferred (on the source
            blockchain).
        token_amount : Amount
            The amount of tokens to be transferred (an integer value in
            case of the token's smallest subunit, a decimal value in
            case of the token's main unit).
        service_node_bid : BlockchainAddressBidPair or None
            A pair of the address of the chosen service node and the
            service node's chosen bid.
            If none is specified, the cheapest registered service node
            bid for the token transfer is automatically chosen.
        valid_until_buffer : int
            The buffer in seconds added to the current timestamp plus
            the chosen service node bid's execution time.

        """
        source_blockchain: Blockchain
        destination_blockchain: Blockchain
        sender_private_key: PrivateKey
        recipient_address: BlockchainAddress
        source_token_id: TokenId
        token_amount: Amount
        service_node_bid: typing.Optional[BlockchainAddressBidPair] = None
        valid_until_buffer: int = _DEFAULT_VALID_UNTIL_BUFFER

    def transfer_tokens(self,
                        request: TransferTokensRequest) -> ServiceNodeTaskInfo:
        """Transfer tokens from a sender's account on a source
        blockchain to a recipient's account on a (possibly different)
        destination blockchain.

        Parameters
        ----------
        request : TransferTokensRequest
            The request data for a new token transfer.

        Returns
        -------
        ServiceNodeTaskInfo
            Service node-related information of a token transfer.

        Raises
        ------
        TransferInteractorError
            If the token transfer cannot be executed.

        """
        try:
            find_token_addresses_response = \
                TokenInteractor().find_token_addresses(
                    request.source_blockchain, request.destination_blockchain,
                    request.source_token_id)
            token_amount = self.__compute_token_amount(
                request, find_token_addresses_response.source_token_address)
            service_node_address, service_node_bid = \
                self.__retrieve_service_node_bid(request)
            valid_until = self.__compute_valid_until(request, service_node_bid)
            self.__validate_recipient_address(request)
            source_blockchain_client = get_blockchain_client(
                request.source_blockchain)
            if request.source_blockchain is request.destination_blockchain:
                # Single-chain token transfer
                compute_transfer_signature_request = \
                    BlockchainClient.ComputeTransferSignatureRequest(
                        request.sender_private_key, request.recipient_address,
                        find_token_addresses_response.source_token_address,
                        token_amount, service_node_address, service_node_bid,
                        valid_until)
                compute_transfer_signature_response = \
                    source_blockchain_client.compute_transfer_signature(
                        compute_transfer_signature_request)
                sender_address = \
                    compute_transfer_signature_response.sender_address
                sender_nonce = compute_transfer_signature_response.sender_nonce
                signature = compute_transfer_signature_response.signature
            else:
                # Cross-chain token transfer
                compute_transfer_from_signature_request = \
                    BlockchainClient.ComputeTransferFromSignatureRequest(
                        request.destination_blockchain,
                        request.sender_private_key, request.recipient_address,
                        find_token_addresses_response.source_token_address,
                        find_token_addresses_response.destination_token_address,    # noqa: E501
                        token_amount, service_node_address, service_node_bid,
                        valid_until)
                compute_transfer_from_signature_response = \
                    source_blockchain_client.compute_transfer_from_signature(
                        compute_transfer_from_signature_request)
                sender_address = \
                    compute_transfer_from_signature_response.sender_address
                sender_nonce = \
                    compute_transfer_from_signature_response.sender_nonce
                signature = compute_transfer_from_signature_response.signature
            service_node_url = source_blockchain_client.read_service_node_url(
                service_node_address)
            submit_transfer_request = ServiceNodeClient.SubmitTransferRequest(
                service_node_url, request.source_blockchain,
                request.destination_blockchain, sender_address,
                request.recipient_address,
                find_token_addresses_response.source_token_address,
                find_token_addresses_response.destination_token_address,
                token_amount, service_node_bid, sender_nonce, valid_until,
                signature)
            service_node_task_id = ServiceNodeClient().submit_transfer(
                submit_transfer_request)
            return ServiceNodeTaskInfo(service_node_task_id,
                                       service_node_address)
        except TransferInteractorError:
            raise
        except Exception:
            raise TransferInteractorError('unable to execute a token transfer',
                                          request=request)

    def get_token_transfer_status(self, request: TokenTransferStatusRequest) \
            -> TokenTransferStatus:
        """Get the status of a token transfer.

        Parameters
        ----------
        request : TokenTransferStatusRequest
            The request data for the status of a token transfer.

        Returns
        -------
        TokenTransferStatus
            The data of the token transfer status.

        Raises
        ------
        TransferInteractorError
            If the token transfer status cannot be retrieved.

        """
        try:
            service_node_url = get_blockchain_client(
                request.source_blockchain).read_service_node_url(
                    request.service_node_address)
            source_status = ServiceNodeClient().status(
                service_node_url, request.service_node_task_id)
            token_transfer_status = \
                self.__create_token_transfer_status_response(source_status)
            if source_status.status is not ServiceNodeTransferStatus.CONFIRMED:
                return token_transfer_status
            source_transaction_id = source_status.transaction_id
            token_transfer_status.source_transaction_id = source_transaction_id
            token_transfer_status.source_transfer_id = \
                source_status.transfer_id
            destination_transfer_request = \
                BlockchainClient.DestinationTransferRequest(
                    request.source_blockchain, source_transaction_id)
            try:
                destination_response = get_blockchain_client(
                    source_status.destination_blockchain
                ).read_destination_transfer(destination_transfer_request)
            except UnknownTransferError:
                return token_transfer_status
            token_transfer_status.destination_transfer_status = \
                self.__get_destination_transfer_status(
                    destination_response.latest_block_number,
                    destination_response.transaction_block_number,
                    source_status.destination_blockchain)
            token_transfer_status.destination_transaction_id = \
                destination_response.destination_transaction_id
            token_transfer_status.destination_transfer_id = \
                destination_response.destination_transfer_id
            token_transfer_status.validator_nonce = \
                destination_response.validator_nonce
            token_transfer_status.signer_addresses = \
                destination_response.signer_addresses
            token_transfer_status.signatures = destination_response.signatures
            return token_transfer_status
        except Exception:
            raise TransferInteractorError(
                'unable to get token transfer status', request=request)

    def __compute_token_amount(self, request: TransferTokensRequest,
                               source_token_address: BlockchainAddress) -> int:
        if isinstance(request.token_amount, int):
            return request.token_amount
        return TokenInteractor().convert_amount_to_subunit(
            request.source_blockchain, source_token_address,
            request.token_amount)

    def __compute_valid_until(self, request: TransferTokensRequest,
                              service_node_bid: ServiceNodeBid) -> int:
        if request.valid_until_buffer < 0:
            raise TransferInteractorError(
                '"valid until" buffer must be non-negative',
                valid_until_buffer=request.valid_until_buffer)
        return (math.ceil(time.time()) + service_node_bid.execution_time +
                request.valid_until_buffer)

    def __retrieve_service_node_bid(
            self, request: TransferTokensRequest) \
            -> typing.Tuple[BlockchainAddress, ServiceNodeBid]:
        if request.service_node_bid is None:
            find_cheapest_bid_response = \
                BidInteractor().find_cheapest_service_node_bid(
                    request.source_blockchain, request.destination_blockchain)
            service_node_address = \
                find_cheapest_bid_response.service_node_address
            service_node_bid = find_cheapest_bid_response.service_node_bid
        else:
            service_node_address = request.service_node_bid[0]
            service_node_bid = request.service_node_bid[1]
        return service_node_address, service_node_bid

    def __validate_recipient_address(self, request: TransferTokensRequest):
        recipient_address = request.recipient_address
        source_blockchain_client = get_blockchain_client(
            request.source_blockchain)
        if not source_blockchain_client.is_valid_recipient_address(
                recipient_address):
            raise TransferInteractorError('invalid recipient address',
                                          recipient_address=recipient_address)

    def __create_token_transfer_status_response(
            self,
            transfer_status: ServiceNodeClient.TransferStatusResponse) \
            -> TokenTransferStatus:
        return TokenTransferStatus(
            destination_blockchain=transfer_status.destination_blockchain,
            source_transfer_status=transfer_status.status,
            destination_transfer_status=DestinationTransferStatus.UNKNOWN,
            sender_address=transfer_status.sender_address,
            recipient_address=transfer_status.recipient_address,
            source_token_address=transfer_status.source_token_address,
            destination_token_address=transfer_status.
            destination_token_address, amount=transfer_status.token_amount)

    def __get_destination_transfer_status(
            self, latest_block_number: int, transaction_block_number: int,
            blockchain: Blockchain) -> DestinationTransferStatus:
        confirmations = get_blockchain_config(blockchain)['confirmations']
        if latest_block_number - transaction_block_number < confirmations:
            return DestinationTransferStatus.SUBMITTED
        return DestinationTransferStatus.CONFIRMED



"""Business logic for handling Pantos-compatible tokens.

"""
import dataclasses
import decimal

from pantos.common.blockchains.base import Blockchain
from pantos.common.types import AccountId
from pantos.common.types import Amount
from pantos.common.types import BlockchainAddress
from pantos.common.types import TokenId
from pantos.common.types import TokenSymbol

from pantos.client.library.blockchains import get_blockchain_client
from pantos.client.library.business.base import Interactor
from pantos.client.library.business.base import InteractorError
from pantos.client.library.configuration import get_blockchain_config


class TokenInteractorError(InteractorError):
    """Exception class for all token interactor errors.

    """
    pass


class TokenInteractor(Interactor):
    """Interactor for handling Pantos-compatible tokens.

    """
    def convert_amount_to_main_unit(self, blockchain: Blockchain,
                                    token_id: TokenId,
                                    amount_subunit: int) -> decimal.Decimal:
        """Convert an amount from a token's smallest subunit to its main
        unit.

        Parameters
        ----------
        blockchain : Blockchain
            The blockchain to convert the token amount for.
        token_id : TokenId
            The identifier of the token.
        amount_subunit : int
            The amount in the token's smallest subunit.

        Returns
        -------
        decimal.Decimal
            The amount in the token's main unit.

        Raises
        ------
        TokenInteractorError
            If the amount cannot be converted.

        """
        try:
            if amount_subunit < 0:
                raise TokenInteractorError('amount must be non-negative',
                                           amount_subunit=amount_subunit)
            if amount_subunit == 0:
                return decimal.Decimal(0)
            token_address = self.__token_id_to_token_address(
                blockchain, token_id)
            blockchain_client = get_blockchain_client(blockchain)
            token_decimals = blockchain_client.read_token_decimals(
                token_address)
            assert token_decimals >= 0
            return decimal.Decimal(amount_subunit) / (10**token_decimals)
        except TokenInteractorError:
            raise
        except Exception:
            raise TokenInteractorError(
                'unable to convert an amount to a token\'s main unit',
                blockchain=blockchain, token_id=token_id,
                amount_subunit=amount_subunit)

    def convert_amount_to_subunit(self, blockchain: Blockchain,
                                  token_id: TokenId,
                                  amount_main_unit: decimal.Decimal) -> int:
        """Convert an amount from a token's main unit to its smallest
        subunit.

        Parameters
        ----------
        blockchain : Blockchain
            The blockchain to convert the token amount for.
        token_id : TokenId
            The identifier of the token.
        amount_main_unit : decimal.Decimal
            The amount in the token's main unit.

        Returns
        -------
        int
            The amount in the token's smallest subunit.

        Raises
        ------
        TokenInteractorError
            If the amount cannot be converted.

        """
        try:
            if amount_main_unit < 0:
                raise TokenInteractorError('amount must be non-negative',
                                           amount_main_unit=amount_main_unit)
            if amount_main_unit == 0:
                return 0
            token_address = self.__token_id_to_token_address(
                blockchain, token_id)
            blockchain_client = get_blockchain_client(blockchain)
            token_decimals = blockchain_client.read_token_decimals(
                token_address)
            assert token_decimals >= 0
            amount_subunit = amount_main_unit * (10**token_decimals)
            amount_subunit_integer = int(amount_subunit)
            if (amount_subunit - amount_subunit_integer) != 0:
                raise TokenInteractorError(
                    'amount must not have more decimals than token',
                    amount_main_unit=amount_main_unit,
                    token_decimals=token_decimals)
            return amount_subunit_integer
        except TokenInteractorError:
            raise
        except Exception:
            raise TokenInteractorError(
                'unable to convert an amount to a token\'s smallest subunit',
                blockchain=blockchain, token_id=token_id,
                amount_main_unit=amount_main_unit)

    def find_token_address(self, blockchain: Blockchain,
                           token_symbol: TokenSymbol) -> BlockchainAddress:
        """Find the blockchain address of a token by its symbol.

        Parameters
        ----------
        blockchain : Blockchain
            The blockchain to find the token address on.
        token_symbol : TokenSymbol
            The symbol of the token.

        Returns
        -------
        BlockchainAddress
            The found token address.

        Raises
        ------
        TokenInteractorError
            If the token symbol is unknown.

        """
        blockchain_config = get_blockchain_config(blockchain)
        token_address = blockchain_config['tokens'].get(token_symbol.lower())
        if token_address is None:
            raise TokenInteractorError('token symbol unknown',
                                       blockchain=blockchain,
                                       token_symbol=token_symbol)
        return BlockchainAddress(token_address)

    @dataclasses.dataclass
    class FindTokenAddressesResponse:
        """Response data for finding blockchain addresses of a token on
        a source and destination blockchain.

        Attributes
        ----------
        source_token_address : BlockchainAddress
            The token's address on the source blockchain.
        destination_token_address : BlockchainAddress
            The token's address on the destination blockchain.

        """
        source_token_address: BlockchainAddress
        destination_token_address: BlockchainAddress

    def find_token_addresses(
            self, source_blockchain: Blockchain,
            destination_blockchain: Blockchain,
            source_token_id: TokenId) -> FindTokenAddressesResponse:
        """Find the blockchain addresses of a token on a source and
        destination blockchain.

        Parameters
        ----------
        source_blockchain : Blockchain
            The source blockchain to find the token address on.
        destination_blockchain : Blockchain
            The destination blockchain to find the token address on.
        source_token_id : TokenId
            The identifier of the token on the source blockchain.

        Returns
        -------
        FindTokenAddressesResponse
            The response data containing the source and destination
            token addresses.

        Raises
        ------
        TokenInteractorError
            If the token identifier is unknown or the token addresses
            cannot be searched for.

        """
        try:
            source_token_address = self.__token_id_to_token_address(
                source_blockchain, source_token_id)
            if source_blockchain is destination_blockchain:
                destination_token_address = source_token_address
            elif isinstance(source_token_id, TokenSymbol):
                destination_token_address = self.find_token_address(
                    destination_blockchain, source_token_id)
            else:
                source_blockchain_client = get_blockchain_client(
                    source_blockchain)
                destination_token_address = \
                    source_blockchain_client.read_external_token_address(
                        source_token_address, destination_blockchain)
            return TokenInteractor.FindTokenAddressesResponse(
                source_token_address, destination_token_address)
        except TokenInteractorError:
            raise
        except Exception:
            raise TokenInteractorError(
                'unable to search for token addresses',
                source_blockchain=source_blockchain,
                destination_blockchain=destination_blockchain,
                source_token_id=source_token_id)

    @dataclasses.dataclass
    class RetrieveTokenBalanceRequest:
        """Request data for retrieving the token balance of a blockchain
        account.

        Attributes
        ----------
        blockchain : Blockchain
            The blockchain to retrieve the token balance on.
        token_id : TokenId
            The identifier of the token.
        account_id : AccountId
            The identifier of the blockchain account.
        return_in_main_unit : bool
            True if the token balance is to be returned in the token's
            main unit, False if it is to be returned in the token's
            smallest subunit.

        """
        blockchain: Blockchain
        token_id: TokenId
        account_id: AccountId
        return_in_main_unit: bool

    def retrieve_token_balance(self,
                               request: RetrieveTokenBalanceRequest) -> Amount:
        """Retrieve the token balance of a blockchain account.

        Parameters
        ----------
        request : RetrieveTokenBalanceRequest
            The request data for retrieving the token balance.

        Returns
        -------
        Amount
            The token balance of the blockchain account (an integer
            value in case of the token's smallest subunit, a decimal
            value in case of the token's main unit).

        Raises
        ------
        TokenInteractorError
            If the token balance cannot be retrieved.

        """
        try:
            token_address = self.__token_id_to_token_address(
                request.blockchain, request.token_id)
            blockchain_client = get_blockchain_client(request.blockchain)
            token_balance = blockchain_client.read_token_balance(
                token_address, request.account_id)
            assert token_balance >= 0
            return (token_balance if not request.return_in_main_unit
                    else self.convert_amount_to_main_unit(
                        request.blockchain, token_address, token_balance))
        except TokenInteractorError:
            raise
        except Exception:
            raise TokenInteractorError(
                'unable to retrieve the token balance of a blockchain account',
                request=request)

    def __token_id_to_token_address(self, blockchain: Blockchain,
                                    token_id: TokenId) -> BlockchainAddress:
        if isinstance(token_id, BlockchainAddress):
            return token_id
        return self.find_token_address(blockchain, token_id)



"""Base classes for all blockchain clients and errors.

"""
import abc
import dataclasses
import typing

import semantic_version  # type: ignore
from pantos.common.blockchains.base import Blockchain
from pantos.common.blockchains.base import BlockchainHandler
from pantos.common.blockchains.base import BlockchainUtilities
from pantos.common.blockchains.base import BlockchainUtilitiesError
from pantos.common.blockchains.factory import get_blockchain_utilities
from pantos.common.blockchains.factory import initialize_blockchain_utilities
from pantos.common.entities import ServiceNodeBid
from pantos.common.exceptions import ErrorCreator
from pantos.common.types import AccountId
from pantos.common.types import BlockchainAddress
from pantos.common.types import PrivateKey

from pantos.client.library.configuration import get_blockchain_config
from pantos.client.library.exceptions import ClientLibraryError
from pantos.client.library.protocol import is_supported_protocol_version


class BlockchainClientError(ClientLibraryError):
    """Base exception class for all blockchain client errors.

    """
    pass


class UnknownTransferError(BlockchainClientError):
    """Exception raised for unknown token transfers.

    """
    def __init__(self, **kwargs: typing.Any):
        # Docstring inherited
        super().__init__('unknown transfer', **kwargs)


class BlockchainClient(BlockchainHandler, ErrorCreator[BlockchainClientError]):
    """Base class for all blockchain clients.

    Attributes
    ----------
    protocol_version : semantic_version.Version
        The version of the Pantos protocol that the blockchain client
        instance is compliant with.

    """
    def __init__(self, protocol_version: semantic_version.Version):
        """Construct a blockchain client instance.

        Parameters
        ----------
        protocol_version : semantic_version.Version
            The version of the Pantos protocol that the blockchain
            client instance must comply with.

        Raises
        ------
        BlockchainClientError
            If the blockchain is not active in the configuration or if
            the corresponding blockchain utilities cannot be
            initialized.

        """
        if not self._get_config()['active']:
            raise self._create_error('blockchain is not active')
        assert is_supported_protocol_version(protocol_version)
        self.protocol_version: typing.Final[
            semantic_version.Version] = protocol_version
        blockchain_node_url = self._get_config()['provider']
        fallback_blockchain_nodes_urls = self._get_config().get(
            'fallback_providers', [])
        average_block_time = self._get_config()['average_block_time']
        required_transaction_confirmations = \
            self._get_config()['confirmations']
        transaction_network_id = self._get_config().get('chain_id')
        try:
            initialize_blockchain_utilities(
                self.get_blockchain(), [blockchain_node_url],
                fallback_blockchain_nodes_urls, average_block_time,
                required_transaction_confirmations, transaction_network_id)
        except BlockchainUtilitiesError:
            raise self._create_error(
                'unable to initialize the {} utilities'.format(
                    self.get_blockchain_name()))

    @dataclasses.dataclass
    class ComputeTransferSignatureRequest:
        """Request data for computing the sender's signature for a
        single-chain token transfer.

        Attributes
        ----------
        sender_private_key : PrivateKey
            The unencrypted private key of the sender's account.
        recipient_address : BlockchainAddress
            The address of the recipient's account.
        token_address : BlockchainAddress
            The transferred token's blockchain address.
        token_amount : int
            The transferred token amount (in 10^-d units, where d is the
            token's number of decimals).
        service_node_address : BlockchainAddress
            The address of the service node used for the token transfer.
        service_node_bid : ServiceNodeBid
            The service node bid used for the token transfer.
        valid_until : int
            The timestamp until when the token transfer is valid (in
            seconds since the epoch).

        """
        sender_private_key: PrivateKey
        recipient_address: BlockchainAddress
        token_address: BlockchainAddress
        token_amount: int
        service_node_address: BlockchainAddress
        service_node_bid: ServiceNodeBid
        valid_until: int

    @dataclasses.dataclass
    class ComputeTransferSignatureResponse:
        """Response data for computing the sender's signature for a
        single-chain token transfer.

        Attributes
        ----------
        sender_address : BlockchainAddress
            The address of the sender's account.
        sender_nonce : int
            The unique nonce of the sender for the token transfer.
        signature : str
            The sender's token transfer signature.

        """
        sender_address: BlockchainAddress
        sender_nonce: int
        signature: str

    @dataclasses.dataclass
    class DestinationTransferRequest:
        """Request data for a token transfer on the destination
        blockchain.

        Attributes
        ----------
        source_blockchain : Blockchain
            The token transfer's source blockchain.
        source_transaction_id : str
            The transaction ID of the token transfer on the
            source blockchain.
        blocks_to_search : int | None
            The blocks to search for the token transfer on the
            destination blockchain.

        """
        source_blockchain: Blockchain
        source_transaction_id: str
        blocks_to_search: int | None = None

    @dataclasses.dataclass
    class DestinationTransferResponse:
        """Response data for a token transfer on the destination
        blockchain.

        Attributes
        ----------
        latest_block_number : int
            The latest block number on the destination blockchain.
        transaction_block_number : int
            The block number of the token transfer transaction.
        destination_transaction_id : str
            The transaction ID of the token transfer.
        source_transfer_id : int
            The unique identifier of the token transfer on the source
            blockchain.
        destination_transfer_id : int
            The unique identifier of the token transfer on the
            destination blockchain.
        sender_address : BlockchainAddress
            The address of the sender's account.
        recipient_address : BlockchainAddress
            The address of the recipient's account.
        source_token_address : BlockchainAddress
            The transferred token's address on the source blockchain.
        destination_token_address : BlockchainAddress
            The transferred token's address on the destination
            blockchain.
        amount : int
            The transferred token amount.
        validator_nonce : int
            The unique nonce of the validator for the token transfer on
            the destination blockchain.
        signer_addresses : list of BlockchainAddress
            The addresses of the validators which signed the token
            transfer on the destination blockchain.
        signatures : list of str
            The signatures of the validators which signed the token
            transfer on the destination blockchain.

        """
        latest_block_number: int
        transaction_block_number: int
        destination_transaction_id: str
        source_transfer_id: int
        destination_transfer_id: int
        sender_address: BlockchainAddress
        recipient_address: BlockchainAddress
        source_token_address: BlockchainAddress
        destination_token_address: BlockchainAddress
        amount: int
        validator_nonce: int
        signer_addresses: list[BlockchainAddress]
        signatures: list[str]

    @abc.abstractmethod
    def compute_transfer_signature(
            self, request: ComputeTransferSignatureRequest) \
            -> ComputeTransferSignatureResponse:
        """Compute the sender's signature for a single-chain token
        transfer.

        Parameters
        ----------
        request : ComputeTransferSignatureRequest
            The request data for computing the transfer signature.

        Returns
        -------
        ComputeTransferSignatureResponse
            The response data with the computed transfer signature.

        Raises
        ------
        BlockchainClientError
            If the signature for the token transfer cannot be computed.

        """
        pass  # pragma: no cover

    @dataclasses.dataclass
    class ComputeTransferFromSignatureRequest:
        """Request data for computing the sender's signature for a
        cross-chain token transfer.

        Attributes
        ----------
        destination_blockchain : Blockchain
            The token transfer's destination blockchain.
        sender_private_key : PrivateKey
            The unencrypted private key of the sender's account on the
            source blockchain.
        recipient_address : BlockchainAddress
            The address of the recipient's account on the destination
            blockchain.
        source_token_address : BlockchainAddress
            The transferred token's address on the source blockchain.
        destination_token_address : BlockchainAddress
            The transferred token's address on the destination
            blockchain.
        token_amount : int
            The transferred token amount (in 10^-d units, where d is the
            token's number of decimals).
        service_node_address : BlockchainAddress
            The address of the service node used for the token transfer.
        service_node_bid : ServiceNodeBid
            The service node bid used for the token transfer.
        valid_until : int
            The timestamp until when the token transfer is valid (in
            seconds since the epoch).

        """
        destination_blockchain: Blockchain
        sender_private_key: PrivateKey
        recipient_address: BlockchainAddress
        source_token_address: BlockchainAddress
        destination_token_address: BlockchainAddress
        token_amount: int
        service_node_address: BlockchainAddress
        service_node_bid: ServiceNodeBid
        valid_until: int

    @dataclasses.dataclass
    class ComputeTransferFromSignatureResponse:
        """Response data for computing the sender's signature for a
        cross-chain token transfer.

        Attributes
        ----------
        sender_address : BlockchainAddress
            The address of the sender's account on the source
            blockchain.
        sender_nonce : int
            The unique nonce of the sender for the token transfer on the
            source blockchain.
        signature : str
            The sender's token transfer signature.

        """
        sender_address: BlockchainAddress
        sender_nonce: int
        signature: str

    @abc.abstractmethod
    def compute_transfer_from_signature(
            self, request: ComputeTransferFromSignatureRequest) \
            -> ComputeTransferFromSignatureResponse:
        """Compute the sender's signature for a cross-chain token
        transfer.

        Parameters
        ----------
        request : ComputeTransferFromSignatureRequest
            The request data for computing the transfer signature.

        Returns
        -------
        ComputeTransferFromSignatureResponse
            The response data with the computed transfer signature.

        Raises
        ------
        BlockchainClientError
            If the signature for the token transfer cannot be computed.

        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def is_valid_recipient_address(self, recipient_address: str) -> bool:
        """Determine if an address string is a valid recipient address
        on the blockchain.

        Parameters
        ----------
        recipient_address : str
            The address string to check.

        Returns
        -------
        bool
            True if the given address string is a valid recipient
            address on the blockchain.

        """
        pass  # pragma: no cover

    def decrypt_private_key(self, keystore: str, password: str) -> PrivateKey:
        """Decrypt the private key from a password-encrypted keystore.

        Parameters
        ----------
        keystore : str
            The keystore contents.
        password : str
            The password to decrypt the private key.

        Returns
        -------
        PrivateKey
            The decrypted private key.

        Raises
        ------
        BlockchainClientError
            If the private key cannot be loaded from the keystore file.

        """
        try:
            return PrivateKey(self._get_utilities().decrypt_private_key(
                keystore, password))
        except Exception:
            raise self._create_error(
                'unable to load a private key from a keystore',
                keystore=keystore)

    @abc.abstractmethod
    def read_external_token_address(
            self, token_address: BlockchainAddress,
            destination_blockchain: Blockchain) -> BlockchainAddress:
        """Read an external token address that is registered at the
        Pantos Hub on the blockchain.

        Parameters
        ----------
        token_address : BlockchainAddress
            The (native) blockchain address of the token.
        destination_blockchain : Blockchain
            The blockchain to read the token's external address for.

        Returns
        -------
        BlockchainAddress
            The external blockchain address of the token.

        Raises
        ------
        BlockchainClientError
            If the external token address cannot be read or if it is not
            active.

        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def read_service_node_addresses(self) -> list[BlockchainAddress]:
        """Read the blockchain addresses of the active service nodes
        registered at the Pantos Hub on the blockchain.

        Returns
        -------
        list of BlockchainAddress
            The blockchain addresses of the active service nodes.

        Raises
        ------
        BlockchainClientError
            If the blockchain adresses of the active service nodes
            cannot be read.

        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def read_service_node_url(self,
                              service_node_address: BlockchainAddress) -> str:
        """Read a service node's URL that is registered at the Pantos
        Hub on the blockchain.

        Parameters
        ----------
        service_node_address : BlockchainAddress
            The blockchain address of the service node.

        Returns
        -------
        str
            The service node's registered URL.

        Raises
        ------
        BlockchainClientError
            If the service node's registered URL cannot be read or if
            the service node is not active.

        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def read_destination_transfer(
            self, request: DestinationTransferRequest) \
            -> DestinationTransferResponse:
        """Read a token transfer on the destination blockchain.

        Parameters
        ----------
        request : DestinationTransferRequest
            The request data for reading the token transfer.

        Returns
        -------
        DestinationTransferResponse
            The response data with the token transfer information.

        Raises
        ------
        UnknownTransferError
            If the token transfer is unkown.
        BlockchainClientError
            If the token transfer cannot be read.

        """
        pass  # pragma: no cover

    def read_token_balance(self, token_address: BlockchainAddress,
                           account_id: AccountId) -> int:
        """Read a blockchain account's balance of a Pantos-compatible
        token.

        Parameters
        ----------
        token_address : BlockchainAddress
            The blockchain address of the token.
        account_id : AccountId
            The identifier of the blockchain account.

        Returns
        -------
        int
            The blockchain account's token balance in the token's
            smallest subunit.

        Raises
        ------
        BlockchainClientError
            If the blockchain account's token balance cannot be read.

        """
        try:
            account_address = self._account_id_to_account_address(account_id)
            return self._get_utilities().get_balance(
                account_address, token_address=token_address)
        except Exception:
            raise self._create_error(
                'unable to read the token balance of a blockchain account',
                token_address=token_address, account_id=account_id)

    @abc.abstractmethod
    def read_token_decimals(self, token_address: BlockchainAddress) -> int:
        """Read the number of decimals of a Pantos-compatible token.

        Parameters
        ----------
        token_address : BlockchainAddress
            The blockchain address of the token.

        Returns
        -------
        int
            The number of decimals of the token.

        Raises
        ------
        BlockchainClientError
            If the token's number of decimals cannot be read.

        """
        pass  # pragma: no cover

    def _create_unknown_transfer_error(
            self, **kwargs: typing.Any) -> BlockchainClientError:
        return self._create_error(specialized_error_class=UnknownTransferError,
                                  **kwargs)

    def _account_id_to_account_address(
            self, account_id: AccountId) -> BlockchainAddress:
        if isinstance(account_id, BlockchainAddress):
            return account_id
        return BlockchainAddress(self._get_utilities().get_address(account_id))

    def _get_config(self) -> dict[str, typing.Any]:
        return get_blockchain_config(self.get_blockchain())

    def _get_utilities(self) -> BlockchainUtilities:
        return get_blockchain_utilities(self.get_blockchain())



"""Module for BNB-Chain-specific clients and errors. Since the BNB
Smart Chain is Ethereum-compatible, the client implementation inherits
from the pantos.client.library.blockchains.ethereum module.

"""
from pantos.common.blockchains.base import Blockchain

from pantos.client.library.blockchains import BlockchainClientError
from pantos.client.library.blockchains.ethereum import EthereumClient
from pantos.client.library.blockchains.ethereum import EthereumClientError


class BnbChainClientError(EthereumClientError):
    """Exception class for all BNB Chain client errors.

    """
    pass


class BnbChainClient(EthereumClient):
    """BNB-Chain-specific blockchain client.

    """
    @classmethod
    def get_blockchain(cls) -> Blockchain:
        # Docstring inherited
        return Blockchain.BNB_CHAIN

    @classmethod
    def get_error_class(cls) -> type[BlockchainClientError]:
        # Docstring inherited
        return BnbChainClientError



"""Module for keeping track of supported Pantos protocol versions.

"""
import typing

import semantic_version  # type: ignore
from pantos.common.protocol import \
    is_supported_protocol_version as _is_supported_protocol_version_by_common

_SUPPORTED_PROTOCOL_VERSIONS: typing.Final[set[semantic_version.Version]] = {
    semantic_version.Version('0.1.0')
}

for supported_protocol_version in _SUPPORTED_PROTOCOL_VERSIONS:
    assert _is_supported_protocol_version_by_common(supported_protocol_version)


def get_latest_protocol_version() -> semantic_version.Version:
    """Get the latest supported Pantos protocol version.

    Returns
    -------
    semantic_version.Version
        The protocol version.

    """
    return max(_SUPPORTED_PROTOCOL_VERSIONS)


def get_supported_protocol_versions() -> list[semantic_version.Version]:
    """Get all supported Pantos protocol versions.

    Returns
    -------
    list of semantic_version.Version
        The sorted protocol versions.

    """
    return sorted(_SUPPORTED_PROTOCOL_VERSIONS)


def is_supported_protocol_version(version: semantic_version.Version) -> bool:
    """Check if a given version is a supported Pantos protocol version.

    Parameters
    ----------
    version : semantic_version.Version
        The version to check.

    Returns
    -------
    bool
        True if the protocol version is supported.

    """
    return version in _SUPPORTED_PROTOCOL_VERSIONS



"""Module that implements the Pantos client library's API which is
exposed to the users of the library.

"""
__all__ = [
    'Blockchain', 'BlockchainAddress', 'PantosClientError', 'PrivateKey',
    'ServiceNodeBid', 'TokenSymbol', 'ServiceNodeTaskInfo',
    'DestinationTransferStatus', 'TokenTransferStatus', 'decrypt_private_key',
    'retrieve_service_node_bids', 'retrieve_token_balance', 'transfer_tokens',
    'get_token_transfer_status', 'deploy_pantos_compatible_token'
]

import uuid as _uuid

from pantos.common.blockchains.base import Blockchain
from pantos.common.entities import \
    BlockchainAddressBidPair as _BlockchainAddressBidPair
from pantos.common.entities import ServiceNodeBid
from pantos.common.types import AccountId as _AccountId
from pantos.common.types import Amount as _Amount
from pantos.common.types import BlockchainAddress
from pantos.common.types import PrivateKey
from pantos.common.types import TokenId as _TokenId
from pantos.common.types import TokenSymbol

from pantos.client.library import initialize_library as _initialize_library
from pantos.client.library.blockchains import \
    get_blockchain_client as _get_blockchain_client
from pantos.client.library.business.bids import BidInteractor as _BidInteractor
from pantos.client.library.business.deployments import \
    TokenDeploymentInteractor as _TokenDeploymentInteractor
from pantos.client.library.business.tokens import \
    TokenInteractor as _TokenInteractor
from pantos.client.library.business.transfers import \
    TransferInteractor as _TransferInteractor
from pantos.client.library.constants import \
    TOKEN_SYMBOL_PAN as _TOKEN_SYMBOL_PAN
from pantos.client.library.entitites import DestinationTransferStatus
from pantos.client.library.entitites import ServiceNodeTaskInfo
from pantos.client.library.entitites import TokenTransferStatus
from pantos.client.library.exceptions import ClientError as _ClientError

# Exception to be used by external client library users
PantosClientError = _ClientError


def decrypt_private_key(blockchain: Blockchain, keystore: str,
                        password: str) -> PrivateKey:
    """Decrypt the private key from a password-encrypted keystore.

    Parameters
    ----------
    blockchain : Blockchain
        The blockchain to load the private key for.
    keystore: str
        The keystore contents.
    password : str
        The password to decrypt the private key.

    Returns
    -------
    PrivateKey
        The decrypted private key.

    Raises
    ------
    PantosClientError
        If the private key cannot be loaded from the keystore file.

    """
    _initialize_library(False)
    return _get_blockchain_client(blockchain).decrypt_private_key(
        keystore, password)


def retrieve_service_node_bids(
        source_blockchain: Blockchain, destination_blockchain: Blockchain,
        return_fee_in_main_unit: bool = True, *, mainnet: bool = False) \
        -> dict[BlockchainAddress, list[ServiceNodeBid]]:
    """Retrieve the service node bids for token transfers from a
    specified source blockchain to a specified destination blockchain.

    Parameters
    ----------
    source_blockchain : Blockchain
        The source blockchain of the service node bids.
    destination_blockchain : Blockchain
        The destination blockchain of the service node bids.
    return_fee_in_main_unit : bool, optional
        True if the service node bids' fee is to be returned in the
        Pantos Token's main unit, False if it is to be returned in the
        Pantos Token's smallest subunit (default: True).
    mainnet : bool, optional
        If True, the function is executed on mainnet. Otherwise, it is
        executed on testnet (default: testnet).

    Returns
    -------
    dict of BlockchainAddress and list of ServiceNodeBid
        The matching service node bids of each registered service
        node.

    Raises
    ------
    PantosClientError
        If the service node bids cannot be retrieved.

    """
    _initialize_library(mainnet)
    return _BidInteractor().retrieve_service_node_bids(
        source_blockchain, destination_blockchain, return_fee_in_main_unit)


def retrieve_token_balance(blockchain: Blockchain, account_id: _AccountId,
                           token_id: _TokenId = _TOKEN_SYMBOL_PAN,
                           return_in_main_unit: bool = True, *,
                           mainnet: bool = False) -> _Amount:
    """Retrieve the token balance of a blockchain account.

    Parameters
    ----------
    blockchain : Blockchain
        The blockchain to retrieve the token balance on.
    account_id : BlockchainAddress or PrivateKey
        The address or unencrypted private key of the blockchain
        account.
    token_id : BlockchainAddress or TokenSymbol
        The address or symbol of the token (default: PAN).
    return_in_main_unit : bool, optional
        True if the token balance is to be returned in the token's main
        unit, False if it is to be returned in the token's smallest
        subunit (default: True).
    mainnet : bool, optional
        If True, the function is executed on mainnet. Otherwise, it is
        executed on testnet (default: testnet).

    Returns
    -------
    int or decimal.Decimal
        The token balance of the blockchain account (an integer value in
        case of the token's smallest subunit, a decimal value in case of
        the token's main unit).

    Raises
    ------
    PantosClientError
        If the token balance cannot be retrieved.

    """
    _initialize_library(mainnet)
    request = _TokenInteractor.RetrieveTokenBalanceRequest(
        blockchain, token_id, account_id, return_in_main_unit)
    return _TokenInteractor().retrieve_token_balance(request)


def transfer_tokens(source_blockchain: Blockchain,
                    destination_blockchain: Blockchain,
                    sender_private_key: PrivateKey,
                    recipient_address: BlockchainAddress,
                    source_token_id: _TokenId, token_amount: _Amount,
                    service_node_bid: _BlockchainAddressBidPair | None = None,
                    *, mainnet: bool = False) -> ServiceNodeTaskInfo:
    """Transfer tokens from a sender's account on a source blockchain to
    a recipient's account on a (possibly different) destination blockchain.

    Parameters
    ----------
    source_blockchain : Blockchain
        The token transfer's source blockchain.
    destination_blockchain : Blockchain
        The token transfer's destination blockchain.
    sender_private_key : PrivateKey
        The unencrypted private key of the sender's account on the
        source blockchain.
    recipient_address : BlockchainAddress
        The address of the recipient's account on the destination
        blockchain.
    source_token_id : BlockchainAddress or TokenSymbol
        The address or symbol of the token to be transferred (on the
        source blockchain).
    token_amount : int or decimal.Decimal
        The amount of tokens to be transferred (an integer value in case
        of the token's smallest subunit, a decimal value in case of the
        token's main unit).
    service_node_bid : tuple of ServiceNodeBid and int or None
        A pair of the address of the chosen service node and the
        service node's chosen bid. If none is specified,
        the registered service node bid with the lowest
        fee for the token transfer is automatically chosen.
    mainnet : bool, optional
        If True, the function is executed on mainnet. Otherwise, it is
        executed on testnet (default: testnet).

    Returns
    -------
    ServiceNodeTaskInfo
        Service node-related information of a token transfer.

    Raises
    ------
    PantosClientError
        If the token transfer cannot be executed.

    """
    _initialize_library(mainnet)
    request = _TransferInteractor.TransferTokensRequest(
        source_blockchain, destination_blockchain, sender_private_key,
        recipient_address, source_token_id, token_amount, service_node_bid)
    return _TransferInteractor().transfer_tokens(request)


def get_token_transfer_status(source_blockchain: Blockchain,
                              service_node_address: BlockchainAddress,
                              service_node_task_id: _uuid.UUID,
                              blocks_to_search: int | None = None, *,
                              mainnet: bool = False) -> TokenTransferStatus:
    """Get the status of a token transfer process.

    Parameters
    ----------
    source_blockchain : Blockchain
        The source blockchain of the token transfer.
    service_node_address : BlockchainAddress
        The address of the service node that is handling the token
        transfer.
    service_node_task_id : uuid.UUID
        The service node task ID of the token transfer.
    blocks_to_search : int or None
        The number of blocks to search for the destination transfer.
        If None, the search is performed until the genesis block.
    mainnet : bool, optional
        If True, the function is executed on mainnet. Otherwise, it is
        executed on testnet (default: testnet).

    Returns
    -------
    TokenTransferStatus
        The status of the token transfer transfer.

    Raises
    ------
    PantosClientError
        If the destination transfer cannot be found.

    """
    _initialize_library(mainnet)
    request = _TransferInteractor.TokenTransferStatusRequest(
        source_blockchain, service_node_address, service_node_task_id,
        blocks_to_search)
    return _TransferInteractor().get_token_transfer_status(request)


def deploy_pantos_compatible_token(token_name: str, token_symbol: str,
                                   token_decimals: int, token_pausable: bool,
                                   token_burnable: bool, token_supply: int,
                                   deployment_blockchains: list[Blockchain],
                                   payment_blockchain: Blockchain,
                                   payer_private_key: PrivateKey, *,
                                   mainnet: bool = False) -> _uuid.UUID:
    """Deploy a Pantos-compatible token on the given blockchains.

    Parameters
    ----------
    token_name : str
        The name of the token.
    token_symbol : str
        The symbol of the token.
    token_decimals : int
        The token's number of decimals.
    token_pausable : bool
        If the token is pausable.
    token_burnable : bool
        If the token is burnable.
    token_supply : int
        The supply of the token.
    deployment_blockchains : list of Blockchain
        The blockchains where the deployment will be requested.
    payment_blockchain : Blockchain
        The blockchain on which the payment for the fee will be made.
    payer_private_key : PrivateKey
        The unencrypted private key of the payer's account on the
        payment_blockchain.
    mainnet : bool, optional
        If True, the function is executed on mainnet. Otherwise, it is
        executed on testnet (default: testnet).

    Returns
    -------
    uuid.UUID
        The task ID of the token creator for the deployment process.

    Raises
    ------
    PantosClientError
        If the deployment process cannot be executed.

    """
    _initialize_library(mainnet)
    request = _TokenDeploymentInteractor.TokenDeploymentRequest(
        token_name, token_symbol, token_decimals, token_pausable,
        token_burnable, token_supply, deployment_blockchains,
        payment_blockchain, payer_private_key)
    return _TokenDeploymentInteractor().deploy_token(request)



"""Module for Solana-specific clients and errors.

"""
from pantos.common.blockchains.base import Blockchain
from pantos.common.entities import ServiceNodeBid
from pantos.common.types import BlockchainAddress

from pantos.client.library.blockchains.base import BlockchainClient
from pantos.client.library.blockchains.base import BlockchainClientError


class SolanaClientError(BlockchainClientError):
    """Exception class for all Solana client errors.

    """
    pass


class SolanaClient(BlockchainClient):
    """Solana-specific blockchain client.

    """
    def compute_transfer_signature(
            self, request: BlockchainClient.ComputeTransferSignatureRequest) \
            -> BlockchainClient.ComputeTransferSignatureResponse:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def compute_transfer_from_signature(
            self,
            request: BlockchainClient.ComputeTransferFromSignatureRequest) \
            -> BlockchainClient.ComputeTransferFromSignatureResponse:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def is_valid_recipient_address(self, recipient_address: str) -> bool:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def get_blockchain(cls) -> Blockchain:
        # Docstring inherited
        return Blockchain.SOLANA

    @classmethod
    def get_error_class(cls) -> type[BlockchainClientError]:
        # Docstring inherited
        return SolanaClientError

    def read_external_token_address(
            self, token_address: BlockchainAddress,
            destination_blockchain: Blockchain) -> BlockchainAddress:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def read_service_node_addresses(self) -> list[BlockchainAddress]:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def read_service_node_bid(self, service_node_address: BlockchainAddress,
                              bid_id: int) -> ServiceNodeBid:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def read_service_node_bids(
            self, service_node_address: BlockchainAddress,
            destination_blockchain: Blockchain) -> list[ServiceNodeBid]:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def read_service_node_url(self,
                              service_node_address: BlockchainAddress) -> str:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def read_destination_transfer(
            self, request: BlockchainClient.DestinationTransferRequest) \
            -> BlockchainClient.DestinationTransferResponse:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover

    def read_token_decimals(self, token_address: BlockchainAddress) -> int:
        # Docstring inherited
        raise NotImplementedError  # pragma: no cover



"""Module for Celo-specific clients and errors. Since Celo is
Ethereum-compatible, the client implementation inherits from the
pantos.client.library.blockchains.ethereum module.

"""
from pantos.common.blockchains.base import Blockchain

from pantos.client.library.blockchains import BlockchainClientError
from pantos.client.library.blockchains.ethereum import EthereumClient
from pantos.client.library.blockchains.ethereum import EthereumClientError


class CeloClientError(EthereumClientError):
    """Exception class for all Celo client errors.

    """
    pass


class CeloClient(EthereumClient):
    """Celo-specific blockchain client.

    """
    @classmethod
    def get_blockchain(cls) -> Blockchain:
        # Docstring inherited
        return Blockchain.CELO

    @classmethod
    def get_error_class(cls) -> type[BlockchainClientError]:
        # Docstring inherited
        return CeloClientError



"""Package for all blockchain-specific clients.

"""
__all__ = [
    'BlockchainClient', 'BlockchainClientError', 'get_blockchain_client'
]

from pantos.client.library.blockchains.base import BlockchainClient
from pantos.client.library.blockchains.base import BlockchainClientError
from pantos.client.library.blockchains.factory import get_blockchain_client



"""Common exceptions for the Pantos client.

"""
from pantos.common.exceptions import BaseError


class ClientError(BaseError):
    """Base exception class for all Pantos client errors.

    """
    pass


class ClientLibraryError(ClientError):
    """Base exception class for all Pantos client library errors.

    """
    pass



"""Constants to be used by multiple modules of the Pantos client
library.

"""
import typing

from pantos.common.types import TokenSymbol

TOKEN_SYMBOL_PAN: typing.Final[TokenSymbol] = TokenSymbol('pan')



"""Module for Avalanche-specific clients and errors. Since the
Avalanche C-Chain is Ethereum-compatible, the client implementation
inherits from the pantos.client.library.blockchains.ethereum module.

"""
from pantos.common.blockchains.base import Blockchain

from pantos.client.library.blockchains import BlockchainClientError
from pantos.client.library.blockchains.ethereum import EthereumClient
from pantos.client.library.blockchains.ethereum import EthereumClientError


class AvalancheClientError(EthereumClientError):
    """Exception class for all Avalanche client errors.

    """
    pass


class AvalancheClient(EthereumClient):
    """Avalanche-specific blockchain client.

    """
    @classmethod
    def get_blockchain(cls) -> Blockchain:
        # Docstring inherited
        return Blockchain.AVALANCHE

    @classmethod
    def get_error_class(cls) -> type[BlockchainClientError]:
        # Docstring inherited
        return AvalancheClientError



"""Module for Ethereum-specific clients and errors.

"""
import secrets
import typing

import web3
import web3.contract
import web3.types
from pantos.common.blockchains.base import Blockchain
from pantos.common.blockchains.base import NodeConnections
from pantos.common.blockchains.base import VersionedContractAbi
from pantos.common.blockchains.enums import ContractAbi
from pantos.common.blockchains.ethereum import EthereumUtilities
from pantos.common.types import BlockchainAddress

from pantos.client.library.blockchains.base import BlockchainClient
from pantos.client.library.blockchains.base import BlockchainClientError
from pantos.client.library.blockchains.base import UnknownTransferError
from pantos.client.library.constants import TOKEN_SYMBOL_PAN

_EIP712_DOMAIN_NAME = 'Pantos'

_TRANSFER_MESSAGE_TYPES = {
    'TransferRequest': [{
        'name': 'sender',
        'type': 'address'
    }, {
        'name': 'recipient',
        'type': 'address'
    }, {
        'name': 'token',
        'type': 'address'
    }, {
        'name': 'amount',
        'type': 'uint256'
    }, {
        'name': 'serviceNode',
        'type': 'address'
    }, {
        'name': 'fee',
        'type': 'uint256'
    }, {
        'name': 'nonce',
        'type': 'uint256'
    }, {
        'name': 'validUntil',
        'type': 'uint256'
    }],
    'Transfer': [{
        'name': 'request',
        'type': 'TransferRequest'
    }, {
        'name': 'blockchainId',
        'type': 'uint256'
    }, {
        'name': 'pantosHub',
        'type': 'address'
    }, {
        'name': 'pantosForwarder',
        'type': 'address'
    }, {
        'name': 'pantosToken',
        'type': 'address'
    }]
}

_TRANSFER_FROM_MESSAGE_TYPES = {
    'TransferFromRequest': [{
        'name': 'destinationBlockchainId',
        'type': 'uint256'
    }, {
        'name': 'sender',
        'type': 'address'
    }, {
        'name': 'recipient',
        'type': 'string'
    }, {
        'name': 'sourceToken',
        'type': 'address'
    }, {
        'name': 'destinationToken',
        'type': 'string'
    }, {
        'name': 'amount',
        'type': 'uint256'
    }, {
        'name': 'serviceNode',
        'type': 'address'
    }, {
        'name': 'fee',
        'type': 'uint256'
    }, {
        'name': 'nonce',
        'type': 'uint256'
    }, {
        'name': 'validUntil',
        'type': 'uint256'
    }],
    'TransferFrom': [{
        'name': 'request',
        'type': 'TransferFromRequest'
    }, {
        'name': 'sourceBlockchainId',
        'type': 'uint256'
    }, {
        'name': 'pantosHub',
        'type': 'address'
    }, {
        'name': 'pantosForwarder',
        'type': 'address'
    }, {
        'name': 'pantosToken',
        'type': 'address'
    }]
}

Web3Contract: typing.TypeAlias = NodeConnections.Wrapper[
    web3.contract.Contract]


class EthereumClientError(BlockchainClientError):
    """Exception class for all Ethereum client errors.

    """
    pass


class EthereumClient(BlockchainClient):
    """Ethereum-specific blockchain client.

    """
    def compute_transfer_signature(
            self, request: BlockchainClient.ComputeTransferSignatureRequest) \
            -> BlockchainClient.ComputeTransferSignatureResponse:
        # Docstring inherited
        try:
            sender_address = self._account_id_to_account_address(
                request.sender_private_key)
            node_connections = self._get_utilities().create_node_connections()
            hub_contract = self._create_hub_contract(node_connections)
            sender_nonce = self.__generate_sender_nonce(
                hub_contract, sender_address)
            domain_data = self.__get_eip712_domain_data()
            message_data = self.__get_transfer_message_data(
                request, sender_address, sender_nonce)
            signed_message = web3.Account.sign_typed_data(
                request.sender_private_key, domain_data,
                _TRANSFER_MESSAGE_TYPES, message_data)
            signature = signed_message.signature.to_0x_hex()
            return BlockchainClient.ComputeTransferSignatureResponse(
                sender_address, sender_nonce, signature)
        except Exception:
            raise self._create_error(
                'unable to compute a single-chain transfer signature',
                request=request)

    def compute_transfer_from_signature(
            self,
            request: BlockchainClient.ComputeTransferFromSignatureRequest) \
            -> BlockchainClient.ComputeTransferFromSignatureResponse:
        # Docstring inherited
        try:
            sender_address = self._account_id_to_account_address(
                request.sender_private_key)
            node_connections = self._get_utilities().create_node_connections()
            hub_contract = self._create_hub_contract(node_connections)
            sender_nonce = self.__generate_sender_nonce(
                hub_contract, sender_address)
            domain_data = self.__get_eip712_domain_data()
            message_data = self.__get_transfer_from_message_data(
                request, sender_address, sender_nonce)
            signed_message = web3.Account.sign_typed_data(
                request.sender_private_key, domain_data,
                _TRANSFER_FROM_MESSAGE_TYPES, message_data)
            signature = signed_message.signature.to_0x_hex()
            return BlockchainClient.ComputeTransferFromSignatureResponse(
                sender_address, sender_nonce, signature)
        except Exception:
            raise self._create_error(
                'unable to compute a cross-chain transfer signature',
                request=request)

    def is_valid_recipient_address(self, recipient_address: str) -> bool:
        # Docstring inherited
        is_valid_address = int(recipient_address, 0) != 0 and \
            web3.Web3.is_checksum_address(recipient_address)
        return is_valid_address

    @classmethod
    def get_blockchain(cls) -> Blockchain:
        # Docstring inherited
        return Blockchain.ETHEREUM

    @classmethod
    def get_error_class(cls) -> type[BlockchainClientError]:
        # Docstring inherited
        return EthereumClientError

    def read_external_token_address(
            self, token_address: BlockchainAddress,
            destination_blockchain: Blockchain) -> BlockchainAddress:
        # Docstring inherited
        try:
            node_connections = self._get_utilities().create_node_connections()
            hub_contract = self._create_hub_contract(node_connections)
            external_token_record = \
                hub_contract.caller().getExternalTokenRecord(
                    token_address, destination_blockchain.value).get()
            assert len(external_token_record) == 2
            external_token_active = external_token_record[0]
            if not external_token_active:
                raise self._create_error(
                    'external token is not active',
                    token_address=token_address,
                    destination_blockchain=destination_blockchain)
            external_token_address = external_token_record[1]
            assert isinstance(external_token_address, str)
            return BlockchainAddress(external_token_address)
        except EthereumClientError:
            raise
        except Exception:
            raise self._create_error(
                'unable to read an external token address',
                token_address=token_address,
                destination_blockchain=destination_blockchain)

    def read_service_node_addresses(self) -> list[BlockchainAddress]:
        # Docstring inherited
        try:
            node_connections = self._get_utilities().create_node_connections()
            hub_contract = self._create_hub_contract(node_connections)
            service_node_addresses = \
                hub_contract.caller().getServiceNodes().get()
            return [
                BlockchainAddress(service_node_address)
                for service_node_address in sorted(service_node_addresses)
            ]
        except Exception:
            raise self._create_error(
                'unable to read the active service node addresses')

    def read_service_node_url(self,
                              service_node_address: BlockchainAddress) -> str:
        # Docstring inherited
        try:
            node_connections = self._get_utilities().create_node_connections()
            hub_contract = self._create_hub_contract(node_connections)
            service_node_record = hub_contract.caller().getServiceNodeRecord(
                service_node_address).get()
            assert len(service_node_record) == 5
            service_node_active = service_node_record[0]
            if not service_node_active:
                raise self._create_error(
                    'service node is not active',
                    service_node_address=service_node_address)
            service_node_url = service_node_record[1]
            assert isinstance(service_node_url, str)
            return service_node_url
        except EthereumClientError:
            raise
        except Exception:
            raise self._create_error('unable to read a service node URL',
                                     service_node_address=service_node_address)

    def read_destination_transfer(
            self, request: BlockchainClient.DestinationTransferRequest) \
            -> BlockchainClient.DestinationTransferResponse:
        # Docstring inherited
        try:
            node_connections = self._get_utilities().create_node_connections()
            to_block_number = \
                node_connections.eth.get_block_number().get_minimum_result()
            from_block_number = (to_block_number - request.blocks_to_search +
                                 1 if request.blocks_to_search else 0)
            blocks_per_query = self._get_config()['blocks_per_query']
            hub_contract = self._create_hub_contract(node_connections)
            transfer_event = typing.cast(
                NodeConnections.Wrapper[web3.contract.contract.ContractEvent],
                hub_contract.events.TransferToSucceeded())
            for to_block_number_ in range(to_block_number + 1,
                                          from_block_number,
                                          -blocks_per_query):
                from_block_number_ = max(to_block_number_ - blocks_per_query,
                                         from_block_number)
                transfer_event_logs = self._get_utilities().get_logs(
                    transfer_event, from_block_number_, to_block_number_ - 1)
                transfer_response = self.__find_destination_transfer(
                    transfer_event_logs, request.source_transaction_id,
                    request.source_blockchain, to_block_number)
                if transfer_response:
                    return transfer_response
            raise self._create_unknown_transfer_error(request=request)
        except UnknownTransferError:
            raise
        except Exception:
            raise self._create_error('unable to read a destination transfer',
                                     request=request)

    def read_token_decimals(self, token_address: BlockchainAddress) -> int:
        # Docstring inherited
        try:
            node_connections = self._get_utilities().create_node_connections()
            token_contract = self._create_token_contract(
                node_connections, token_address)
            return token_contract.caller().decimals().get()
        except Exception:
            raise self._create_error(
                'unable to read the number of decimals of a token',
                token_address=token_address)

    def _create_hub_contract(
            self, node_connections: NodeConnections) \
            -> NodeConnections.Wrapper[web3.contract.Contract]:
        try:
            return self._get_utilities().create_contract(
                self._get_config()['hub'],
                VersionedContractAbi(ContractAbi.PANTOS_HUB,
                                     self.protocol_version), node_connections)
        except Exception:
            raise self._create_error(
                'unable to create a hub contract instance')

    def _create_token_contract(
            self, node_connections: NodeConnections,
            token_address: BlockchainAddress) -> Web3Contract:
        try:
            return self._get_utilities().create_contract(
                token_address,
                VersionedContractAbi(ContractAbi.PANTOS_TOKEN,
                                     self.protocol_version), node_connections)
        except Exception:
            raise self._create_error(
                'unable to create a token contract instance')

    def _get_utilities(self) -> EthereumUtilities:
        # Docstring inherited
        return typing.cast(EthereumUtilities, super()._get_utilities())

    def __generate_sender_nonce(self, hub_contract: Web3Contract,
                                sender_address: BlockchainAddress) -> int:
        while True:
            sender_nonce = secrets.randbits(256)
            if hub_contract.caller().isValidSenderNonce(
                    sender_address, sender_nonce).get():
                return sender_nonce

    def __get_eip712_domain_data(self) -> dict[str, typing.Any]:
        return {
            'name': _EIP712_DOMAIN_NAME,
            'version': str(self.protocol_version.major),
            'chainId': self._get_config()['chain_id'],
            'verifyingContract': self._get_config()['forwarder']
        }

    def __get_transfer_message_data(
            self, request: BlockchainClient.ComputeTransferSignatureRequest,
            sender_address: BlockchainAddress,
            sender_nonce: int) -> dict[str, typing.Any]:
        return {
            'request': {
                'sender': sender_address,
                'recipient': request.recipient_address,
                'token': request.token_address,
                'amount': request.token_amount,
                'serviceNode': request.service_node_address,
                'fee': request.service_node_bid.fee,
                'nonce': sender_nonce,
                'validUntil': request.valid_until
            },
            'blockchainId': self.get_blockchain().value,
            'pantosHub': self._get_config()['hub'],
            'pantosForwarder': self._get_config()['forwarder'],
            'pantosToken': self._get_config()['tokens'][TOKEN_SYMBOL_PAN]
        }

    def __get_transfer_from_message_data(
            self,
            request: BlockchainClient.ComputeTransferFromSignatureRequest,
            sender_address: BlockchainAddress,
            sender_nonce: int) -> dict[str, typing.Any]:
        return {
            'request': {
                'destinationBlockchainId': request.destination_blockchain.
                value,
                'sender': sender_address,
                'recipient': request.recipient_address,
                'sourceToken': request.source_token_address,
                'destinationToken': request.destination_token_address,
                'amount': request.token_amount,
                'serviceNode': request.service_node_address,
                'fee': request.service_node_bid.fee,
                'nonce': sender_nonce,
                'validUntil': request.valid_until
            },
            'sourceBlockchainId': self.get_blockchain().value,
            'pantosHub': self._get_config()['hub'],
            'pantosForwarder': self._get_config()['forwarder'],
            'pantosToken': self._get_config()['tokens'][TOKEN_SYMBOL_PAN]
        }

    def __find_destination_transfer(
            self, transfer_event_logs: list[web3.types.EventData],
            source_transaction_id: str, source_blockchain_id: int,
            to_block_number: int) \
            -> BlockchainClient.DestinationTransferResponse | None:
        for transfer_event_log in transfer_event_logs:
            transfer_event_args = transfer_event_log['args']
            transfer_event_request = transfer_event_args['request']
            if (transfer_event_request['sourceTransactionId']
                    == source_transaction_id
                    and transfer_event_request['sourceBlockchainId']
                    == source_blockchain_id):
                return BlockchainClient.DestinationTransferResponse(
                    to_block_number, transfer_event_log['blockNumber'],
                    transfer_event_log['transactionHash'].to_0x_hex(),
                    transfer_event_request['sourceTransferId'],
                    transfer_event_args['destinationTransferId'],
                    BlockchainAddress(transfer_event_request['sender']),
                    BlockchainAddress(transfer_event_request['recipient']),
                    BlockchainAddress(transfer_event_request['sourceToken']),
                    BlockchainAddress(
                        transfer_event_request['destinationToken']),
                    transfer_event_request['amount'],
                    transfer_event_request['nonce'], [
                        BlockchainAddress(signer_address) for signer_address in
                        transfer_event_args['signerAddresses']
                    ], [
                        f"0x{signature.hex()}"
                        for signature in transfer_event_args['signatures']
                    ])
        return None



"""Top-level package of the Pantos client library.

"""
import ctypes as _ctypes
import multiprocessing as _multiprocessing

import semantic_version as _semantic_version  # type: ignore
from pantos.common.configuration import ConfigError as _ConfigError

from pantos.client.library.configuration import config as _config
from pantos.client.library.configuration import load_config as _load_config
from pantos.client.library.exceptions import \
    ClientLibraryError as _ClientLibraryError
from pantos.client.library.protocol import \
    is_supported_protocol_version as _is_supported_protocol_version

_initialized = _multiprocessing.Value(_ctypes.c_bool, False)


def initialize_library(mainnet: bool) -> None:
    """Initialize the Pantos client library. The function is thread-safe
    and performs the initialization only once at the first invocation.

    Parameters
    ----------
    mainnet : bool
        If True, the client library is initialized for mainnet
        operation. Otherwise, it is initialized for testnet operation.

    Raises
    ------
    ClientLibraryError
        If the library cannot be initialized.

    """
    with _initialized.get_lock():
        if not _initialized.value:
            try:
                _load_config()
            except _ConfigError:
                raise _ClientLibraryError('error loading config')
            environment = 'mainnet' if mainnet else 'testnet'
            protocol_version = _semantic_version.Version(
                _config['protocol'][environment])
            if not _is_supported_protocol_version(protocol_version):
                raise _ClientLibraryError(
                    'unsupported Pantos protocol version',
                    protocol_version=protocol_version)
            _initialized.value = True



"""Package for all business logic.

"""
__all__ = ['Interactor', 'InteractorError']

from pantos.client.library.business.base import Interactor
from pantos.client.library.business.base import InteractorError



"""Business logic for handling service node bids.

"""
import concurrent.futures
import dataclasses
import itertools
import secrets
import typing

from pantos.common.blockchains.base import Blockchain
from pantos.common.entities import ServiceNodeBid
from pantos.common.servicenodes import ServiceNodeClient
from pantos.common.types import BlockchainAddress

from pantos.client.library.blockchains import get_blockchain_client
from pantos.client.library.blockchains.base import BlockchainClient
from pantos.client.library.business.base import Interactor
from pantos.client.library.business.base import InteractorError
from pantos.client.library.business.tokens import TokenInteractor
from pantos.client.library.configuration import config
from pantos.client.library.constants import TOKEN_SYMBOL_PAN


class BidInteractorError(InteractorError):
    """Exception class for all bid interactor errors.

    """
    pass


class BidInteractor(Interactor):
    """Interactor for handling service node bids.

    """
    @dataclasses.dataclass
    class CheapestServiceNodeBid:
        """Response data for finding the cheapest service node bid.

        Attributes
        ----------
        service_node_address : BlockchainAddress
            The address of the service node with the cheapest bid.
        service_node_bid: ServiceNodeBid
            The service node bid with the lowest fee. If more than one
            service node bid with the lowest fee is found, the one with
            the lower execution time is selected. If they have the same
            execution time, one is chosen randomly.

        """
        service_node_address: BlockchainAddress
        service_node_bid: ServiceNodeBid

    def find_cheapest_service_node_bid(
            self, source_blockchain: Blockchain,
            destination_blockchain: Blockchain) \
            -> CheapestServiceNodeBid:
        """Find the cheapest service node bid.

        Parameters
        ----------
        source_blockchain : Blockchain
            The source blockchain of the service node bid.
        destination_blockchain : Blockchain
            The destination blockchain of the service node bid.

        Returns
        -------
        CheapestServiceNodeBid
            The response data with the cheapest service node bid.

        Raises
        ------
        BidInteractorError
            If the cheapest service node bid cannot be searched for or
            if no active service node bid is found.

        """
        try:
            cheapest_bid_pairs: typing.List[typing.Tuple[BlockchainAddress,
                                                         ServiceNodeBid]] = []
            all_service_node_bids = self.retrieve_service_node_bids(
                source_blockchain, destination_blockchain, False)
            for service_node_address, service_node_bids in \
                    all_service_node_bids.items():
                if len(service_node_bids) == 0:
                    continue
                bid_pairs = cheapest_bid_pairs + list(
                    zip(itertools.repeat(service_node_address),
                        service_node_bids))
                cheapest_bid = min(
                    bid_pairs, key=lambda bid_pair:
                    (bid_pair[1].fee, bid_pair[1].execution_time))[1]
                cheapest_bid_pairs = [
                    bid_pair for bid_pair in bid_pairs
                    if bid_pair[1].fee == cheapest_bid.fee and
                    bid_pair[1].execution_time == cheapest_bid.execution_time
                ]
            if len(cheapest_bid_pairs) == 0:
                raise BidInteractorError('no active service node bids found')

            service_node_address, service_node_bid = (
                cheapest_bid_pairs[0] if len(cheapest_bid_pairs) == 1 else
                secrets.choice(cheapest_bid_pairs))
            return BidInteractor.CheapestServiceNodeBid(
                service_node_address, service_node_bid)
        except BidInteractorError:
            raise
        except Exception:
            raise BidInteractorError(
                'unable to search for the cheapest service node bid',
                source_blockchain=source_blockchain,
                destination_blockchain=destination_blockchain)

    def retrieve_service_node_bids(
            self, source_blockchain: Blockchain,
            destination_blockchain: Blockchain,
            return_fee_in_main_unit: bool) \
            -> typing.Dict[BlockchainAddress, typing.List[ServiceNodeBid]]:
        """Retrieve the service node bids for token transfers from a
        specified source blockchain to a specified destination
        blockchain.

        Parameters
        ----------
        source_blockchain : Blockchain
            The source blockchain of the service node bids.
        destination_blockchain : Blockchain
            The destination blockchain of the service node bids.
        return_fee_in_main_unit : bool
            True if the service node bids' fee is to be returned in the
            Pantos Token's main unit, False if it is to be returned in
            the Pantos Token's smallest subunit.

        Returns
        -------
        dict of BlockchainAddress and list of ServiceNodeBid
            The matching service node bids of each registered service
            node.

        Raises
        ------
        BidInteractorError
            If the service node bids cannot be retrieved.

        """
        try:
            all_service_node_bids: typing.Dict[
                BlockchainAddress, typing.List[ServiceNodeBid]] = {}
            source_blockchain_client = get_blockchain_client(source_blockchain)
            service_node_addresses = \
                source_blockchain_client.read_service_node_addresses()
            with concurrent.futures.ThreadPoolExecutor() as executor:
                future_to_service_node_address = {
                    executor.submit(
                        self.__retrieve_bid_from_service_node,  # yapf bug
                        source_blockchain_client,
                        destination_blockchain,
                        service_node_address,
                        return_fee_in_main_unit): service_node_address
                    for service_node_address in service_node_addresses
                }
                for future in concurrent.futures.as_completed(
                        future_to_service_node_address):
                    service_node_address = future_to_service_node_address[
                        future]
                    service_node_bids = future.result()
                    if service_node_bids is not None:
                        all_service_node_bids[
                            service_node_address] = service_node_bids
                return all_service_node_bids
        except Exception:
            raise BidInteractorError(
                'unable to retrieve the service node bids',
                source_blockchain=source_blockchain,
                destination_blockchain=destination_blockchain)

    def __retrieve_bid_from_service_node(
            self, source_blockchain_client: BlockchainClient,
            destination_blockchain: Blockchain,
            service_node_address: BlockchainAddress,
            return_fee_in_main_unit: bool) \
            -> typing.Optional[list[ServiceNodeBid]]:
        service_node_url = \
            source_blockchain_client.read_service_node_url(
                service_node_address)
        source_blockchain = source_blockchain_client.get_blockchain()
        timeout = config['service_nodes']['timeout']
        try:
            service_node_bids = ServiceNodeClient().bids(
                service_node_url, source_blockchain, destination_blockchain,
                timeout)
        except Exception:
            return None
        if not return_fee_in_main_unit:
            return service_node_bids
        token_interactor = TokenInteractor()
        for service_node_bid in service_node_bids:
            fee = service_node_bid.fee
            assert isinstance(fee, int)
            service_node_bid.fee = \
                token_interactor.convert_amount_to_main_unit(
                    source_blockchain, TOKEN_SYMBOL_PAN, fee)
        return service_node_bids



"""Module that defines the Pantos client library entities.

"""
import dataclasses
import enum
import uuid

from pantos.common.blockchains.enums import Blockchain
from pantos.common.entities import ServiceNodeTransferStatus
from pantos.common.types import BlockchainAddress


class DestinationTransferStatus(enum.IntEnum):
    """Enumeration of the possible destination transfer statuses.

    """
    UNKNOWN = 0
    SUBMITTED = 1
    CONFIRMED = 2


@dataclasses.dataclass
class ServiceNodeTaskInfo:
    """Service node-related information of a token transfer.

    Attributes
    ----------
    task_id : uuid.UUID
        The task ID of the chosen service node for the token
        transfer.
    service_node_address : BlockchainAddress
        The address of the chosen service node for the token
        transfer.

    """
    task_id: uuid.UUID
    service_node_address: BlockchainAddress


@dataclasses.dataclass
class TokenTransferStatus:
    """Data for the status of a token transfer.

    Attributes
    ----------
    destination_blockchain : Blockchain
        The token transfer's destination blockchain.
    source_transfer_status : ServiceNodeTransferStatus
        The status of the token transfer on the source blockchain.
    destination_transfer_status : DestinationTransferStatus
        The status of the token transfer on the destination blockchain.
    source_transaction_id : str or None
        The transaction ID of the token transfer on the source
        blockchain (default: None).
    destination_transaction_id : str or None
        The transaction ID of the token transfer on the destination
        blockchain (default: None).
    source_transfer_id : int or None
        The transfer ID of the token transfer on the source
        blockchain (default: None).
    destination_transfer_id : int or None
        The transfer ID of the token transfer on the destination
        blockchain (default: None).
    sender_address : BlockchainAddress or None
        The address of the sender's account on the source blockchain
        (default: None).
    recipient_address : BlockchainAddress or None
        The address of the recipient's account on the destination
        blockchain (default: None).
    source_token_address : BlockchainAddress or None
        The address of the token on the source blockchain (default:
        None).
    destination_token_address : BlockchainAddress or None
        The address of the token on the destination blockchain
        (default: None).
    amount : int or None
        The amount of tokens transferred (default: None).
    validator_nonce : int or None
        The validator nonce of the token transfer (default: None).
    signer_addresses : list[BlockchainAddress] or None
        The addresses of the signers of the token transfer (default:
        None).
    signatures : list[str] or None
        The signatures of the token transfer (default: None).

    """
    destination_blockchain: Blockchain
    source_transfer_status: ServiceNodeTransferStatus
    destination_transfer_status: DestinationTransferStatus
    source_transaction_id: str | None = None
    destination_transaction_id: str | None = None
    source_transfer_id: int | None = None
    destination_transfer_id: int | None = None
    sender_address: BlockchainAddress | None = None
    recipient_address: BlockchainAddress | None = None
    source_token_address: BlockchainAddress | None = None
    destination_token_address: BlockchainAddress | None = None
    amount: int | None = None
    validator_nonce: int | None = None
    signer_addresses: list[BlockchainAddress] | None = None
    signatures: list[str] | None = None



"""Factory for blockchain clients.

"""
import semantic_version  # type: ignore
from pantos.common.blockchains.base import Blockchain

from pantos.client.library.blockchains.base import BlockchainClient
from pantos.client.library.protocol import get_latest_protocol_version
from pantos.client.library.protocol import is_supported_protocol_version

_blockchain_clients: dict[tuple[Blockchain, semantic_version.Version],
                          BlockchainClient] = {}
"""Blockchain-specific client objects."""

_blockchain_client_classes = BlockchainClient.find_subclasses()
"""Blockchain-specific client classes."""


def get_blockchain_client(
        blockchain: Blockchain,
        protocol_version: semantic_version.Version | None = None) \
        -> BlockchainClient:
    """Factory for blockchain-specific client objects.

    Parameters
    ----------
    blockchain : Blockchain
        The blockchain to get the client instance for.
    protocol_version : semantic_version.Version, optional
        The version of the Pantos protocol that the blockchain client
        instance must comply with (default: most recent supported
        version).

    Returns
    -------
    BlockchainClient
        A blockchain client instance for the specified blockchain.

    """
    if protocol_version is None:
        protocol_version = get_latest_protocol_version()
    assert is_supported_protocol_version(protocol_version)
    blockchain_client = _blockchain_clients.get((blockchain, protocol_version))
    if blockchain_client is None:
        blockchain_client = _blockchain_client_classes[blockchain](
            protocol_version)
        _blockchain_clients[(blockchain, protocol_version)] = blockchain_client
    return blockchain_client
